/* 

QUEUE DEFINITION --------------------------------------
1.0--> Linker Motor
2.0--> Rechter Motor
3.0-->beide
4.0-->drehung
5.0-->Pen (+1 hoch, -1 runter)

PROTOKOLL ---------------------------------------

/ml/setup        Setup der Wifly mit IP

/ml/alive        Lebenszeichen von ML
/ml/q/start      Ansage von Remote, das Es los gehen kann. Anzahl der Elemente wird als Int gesendet
/ml/q/req        Neue Daten anfragen, als Argument der index
/ml/q/res        Antwort der Remote, mit Index, Mode und Value

/ml/mt/speed     Manuell Touch - Geschwindigkeiten für Rechts und Links
/ml/mt/pen       Manuell Touch - Stift

*/

// DEFINE --------------------------------------------
#define wiflySerialType HardwareSerial  // you can change this to either 'HardwareSerial' or 'SoftwareSerial' as you need it
#define debugSerialType Stream          // you can change this to either 'HardwareSerial' or 'SoftwareSerial' as you need it
#define DEBUG_WIFI_RESPONSE 1 //shall we print all the communication with the wifly module to the debugSerial? [0/1]
#define DEBUG_WIFI_STATUS 1   //shall we show status wifly information on the Serial? [0/1]

// INCLUDE --------------------------------------------
#include "Stepper.h"
#include "Queue.h"
#include "Pen.h"
#include <Servo.h>
#include <MinimalWiflyConfig.h>
#include <ArdOSCForSerial.h>
#include <MemoryFree.h>
#include "TimerThree.h"

// KONSTANTEN -----------------------------------------
int rad_abstand=265;  //Radabstand in mm

// VARIABLEN ----------------------------------------------
boolean queue_is_pause_=false;

// OBJEKTE --------------------------------------------
Stepper stepperLeft;
Stepper stepperRight;
Pen pen;
Queue queue;

unsigned long networkBaudrate=115200;      // Speed of the Serial connection to the network adapter. 19200 is the maximum that works reliably with a Software Serial. 
unsigned long networkTimeoutMillis=3;      // How long the Network adapter waits until it flushes the Buffer and sends an UDP-Package.
OSCClient client(&Serial3,networkTimeoutMillis); // this "OSCClient" object will wrap up the data into an OSC Message and pass it to the netowrk interface
OSCServer server(&Serial3,networkTimeoutMillis);   // the OSC Server will get it's data from the Serial passed to it.
MinimalWiflyConfig wifly (Serial3, Serial);

// SETUP ----------------------------------------------
void setup(){
  
  Serial.begin(115200);
  Serial3.begin(115200);
  
  stepperLeft.pinSetup(22,24);
  stepperRight.pinSetup(26,28);
  
  stepperLeft.setup();
  stepperRight.setup();
  
  stepperLeft.setVMax(300.0);
  stepperRight.setVMax(300.0);
  
  pen.setup();
  queue.setup();
  
  Serial.println("Freier Ram: "+String(freeMemory())+" Byte");

  //Serial.println("AP mode setup");
//  wifly.setAPMode();

  server.addCallback("/ml/q/res",&getQueueResponseTunnel);
  server.addCallback("/ml/q/start",&getQueueStartTunnel);
  server.addCallback("/ml/q/pause",&getQueuePauseTunnel);  
  server.addCallback("/ml/setup",&getSetupTunnel);
 // server.addCallback("/ml/m",&getManuellTunnel);
  server.addCallback("/ml/mt/speed",&getManuellTouchSpeedTunnel);
  server.addCallback("/ml/mt/pen",&getManuellTouchPenTunnel);

 // Timer3.initialize(100);
 // Timer3.attachInterrupt(loopTunnel);
  
}

void getSetupTunnel(OSCMessage *_mes) { // nachricht von remote

  pen.up();
  stepperLeft.setup();
  stepperRight.setup();
  queue.getQueueReset();
  
  int new_ip_1=_mes->getArgInt32(0);
  int new_ip_2=_mes->getArgInt32(1);
  int new_ip_3=_mes->getArgInt32(2);
  int new_ip_4=_mes->getArgInt32(3);
  String ip = String(new_ip_1)+"."+String(new_ip_2)+"."+String(new_ip_3)+"."+String(new_ip_4);

  char *ip_char = new char[ip.length()+1];
  strcpy(ip_char, ip.c_str());
  
  Serial.println(F("Starting Wifly Setup..."));
  // this setup function has to be run only once for a module, because it will remember the settings afterwards.
  // the commands used for configuration are fairly simple - they can be found in the datasheet of the module or in the souce of the library.
  bool success =wifly.setup(
    115200,      // if you use a hardware serial, I would recommend the full 115200. 19200 is the maximum that works reliably with a software serial.
    true,	 // should we set the Module to the specified baudrate if it is currently configured othewise
    "wally",     // Your Wifi Name (SSID)
    "robomint",  // Your Wifi Password 
    "malokal",   // Device name for identification in the network
    0,           // IP Adress of the Wifly. if 0 (without quotes), it will use dhcp to get an ip. Incoming packets must go to the IP adress shown during setup.
    9000,        // WiFly receive port -  incoming packets have to be sent there
    ip_char,     // Where to send outgoing Osc messages. "255.255.255.255" will send to all hosts in the subnet
    9001,        // outgoing port - outgoing packets will be sent there
    MinimalWiflyConfig::PROTO_UDP //protocol bit (see datasheet or library source code)
  );
  Serial.print(F("Wifly setup finished - successful (0=no, 1=yes): "));
  Serial.println(success);

}

// LOOP -----------------------------------------------
void loop(){ 
  
  // nach neuen osc schauen
  server.availableCheck();
  queue.loop();
  
  if (queue.getLength()>0) { //Wenn Queue nicht leer 

    float current_entry_mode=queue.getCurrentEntry(0); // aktueller queue eintrag
    float current_entry_data=queue.getCurrentEntry(1);
  
    // LINKER MOTOR ------------
    if (stepperLeft.getIsReady() && current_entry_mode==1.0) { //wenn er nicht beschäftigt ist und in Queue angesprochen wurde
      stepperLeft.drive(current_entry_data); // entfernung in mm
    }
  
    // RECHTER MOTOR ------------
    if (stepperRight.getIsReady() && current_entry_mode==2.0) {
      stepperRight.drive(current_entry_data);
    }
  
    // BEIDE MOTOREN ------------
    if (stepperLeft.getIsReady() && stepperRight.getIsReady() && current_entry_mode==3.0) {
      stepperLeft.drive(current_entry_data);
       stepperRight.drive(current_entry_data);
    }
  
    // DREHEN ------------
    // Drehwinkel: nach rechts ist 0-180°, nach links -180-0°
    if (stepperLeft.getIsReady() && stepperRight.getIsReady() && current_entry_mode==4.0){
      float dreh_Distanz=PI*rad_abstand*((current_entry_data)/360.0);    //Berechnung der zu fahrenden Distanz aus dem Winkel
     stepperRight.drive(dreh_Distanz*(-1));
     stepperLeft.drive(dreh_Distanz);
    }
  
    // STIFT ------------
    if (current_entry_mode==5.0 && pen.getIsReady()) {
      if (current_entry_data==1.0) {
        pen.up();
      } else {
        pen.down();
      }
    }
    
    stepperLeft.loop();
    stepperRight.loop();
    pen.loop();

    if( // wenn das element des aktuellen mode fertig ist, dann zum nächsten schritt
      (
        (current_entry_mode==1.0 && stepperLeft.getIsReady())||
        (current_entry_mode==2.0 && stepperRight.getIsReady())||
        ((current_entry_mode==3.0 || current_entry_mode==4.0) && stepperLeft.getIsReady() && stepperRight.getIsReady())||
        (current_entry_mode==5.0 && pen.getIsReady())
      )
    )
    {
      queue.removeFromQueue();
    }
      
  } else { // keine ausführung
    queue.setQueueIsRunningFalse();
    
    stepperLeft.loop();
    stepperRight.loop();
    pen.loop();
    
  }
}

// OSC --------------------------------------------

void getQueueResponseTunnel(OSCMessage *_mes) {
  queue.getQueueResponse(_mes);
}

void getQueueStartTunnel(OSCMessage *_mes) {
  pen.up();
  stepperLeft.setup();
  stepperRight.setup();
  queue.getQueueStart(_mes);
}

void getQueuePauseTunnel(OSCMessage *_mes) {   
  if (queue_is_pause_) {
    queue_is_pause_=false;
  } else {
    queue_is_pause_=true;
  }
  
  queue.setQueuePause(queue_is_pause_);
}

/*
void getManuellTunnel(OSCMessage *_mes) {   
    
    float manuell_mode=_mes->getArgFloat(0);
    float manuell_value=_mes->getArgFloat(1);
    
    // 3 geradeaus +1 vorwärts, -1 rückwärts, 0 anhalten
    
    if (manuell_mode==3.0) {
      
      if (manuell_value==1.0) {
         stepperLeft.setManuellOn(1.0); 
         stepperRight.setManuellOn(1.0); 
      }
      
      if (manuell_value==-1.0) {
         stepperLeft.setManuellOn(-1.0); 
         stepperRight.setManuellOn(-1.0); 
      }

      if (manuell_value==0.0) {
         stepperLeft.setManuellOff(); 
         stepperRight.setManuellOff(); 
      }
      
    }
    
    if (manuell_mode==4.0) {
      if (manuell_value==1.0) {
         stepperLeft.setManuellOn(-1.0); 
         stepperRight.setManuellOn(1.0); 
      }
      
      if (manuell_value==-1.0) {
         stepperLeft.setManuellOn(1.0); 
         stepperRight.setManuellOn(-1.0); 
      }

      if (manuell_value==0.0) {
         stepperLeft.setManuellOff(); 
         stepperRight.setManuellOff(); 
      }
    }

    if (manuell_mode==5.0) {
      pen.toggle();

    }
    
}*/

void getManuellTouchSpeedTunnel(OSCMessage *_mes) {   
  float speed_left=_mes->getArgFloat(0);
  float speed_right=_mes->getArgFloat(1);  
  stepperLeft.setManuellV(speed_left);
  stepperRight.setManuellV(speed_right);
  Serial.println("manuell touch "+String(speed_left)+" "+String(speed_right));
}

void getManuellTouchPenTunnel(OSCMessage *_mes) {   
  boolean should_be_up=(int)(_mes->getArgFloat(0))==1;
  pen.setManuell(should_be_up);
  Serial.println("manuell pen "+should_be_up);
}

void sendOSCMessage(OSCMessage* _mes) {
  client.send(_mes);  
}

// INTERUPTS --------------------------------------------

//void loopTunnel() {
   // stepperLeft.loopCalledByTimer();
   // stepperRight.loopCalledByTimer();
//}
