

// INCLUDE --------------------------------------------
#include <MemoryFree.h>
#include <ArdOSCForSerial.h>

// FUNKTIONEN -----------------------------------------
void sendOSCMessage(OSCMessage* _mes);

// KONSTANTEN -----------------------------------------
int queue_length=100; // muss für array seperat definiert werden

// KLASSE ---------------------------------------------
class Queue{

private:

  float queue[100][2];
  int queue_current_index;   //Eintrag, der grade läuft
  int queue_newest_index;    //Letzter angefügter Eintrag

  unsigned long lastMillis;    //Zeitl. Variablen, Zur Delay-freien Verarbeitung
  unsigned long lastMillisAlive;
  unsigned long currentMillis;  //Zeitl. Variablen, seit Programmstart vergangenen Microsekunden

  unsigned long remote_index;
  unsigned long remote_length; // Anzahl an Elementen im aktuellen Vorgang, bei 0 gibt es nichts
  
  boolean queue_is_pause; // gibt an, ob queue gerade angehalten ist
  boolean queue_is_running; // gibt an, ob gerade etwas ausgeführt wird
  
public:

  void setup(){
        
    queue_current_index=0;   // Bei 0 in der Liste anfangen
    queue_newest_index=0;
    lastMillis=0;            // millis 0 seten
    lastMillisAlive=0;       // millis 0 seten

    remote_index=0;
    remote_length=0;
    
    queue_is_pause=false;
    queue_is_running=false;

  }

  void loop() {

    // alle n ms
    currentMillis=millis();
    
    if(remote_length>0&&currentMillis-lastMillis>=300){ // wenn remote verfügbar und zeit rum
      lastMillis=currentMillis;

      if (remote_index<remote_length) { // es gibt noch daten zu holen
        if (getLength()+1<queue_length) { // wenn noch mindestens ein platz frei ist
          sendQueueRequest(); // neue daten holen
          
          if (!queue_is_running) {
            queue_is_running=true;
          }
          
        }
      } else { // Alle Daten geholt
        remote_length=0;
      }
    } 
    
   if(currentMillis-lastMillisAlive>=3000){ // alle 3 sekunden
      lastMillisAlive=currentMillis;
   
      // nachricht an remote, dass ich noch lebe
      OSCMessage local_mes;
      local_mes.beginMessage("/ml/alive");
      local_mes.addArgInt32(queue_current_index);
      local_mes.addArgInt32((int)queue_is_pause); // true=1, false=0      
      local_mes.addArgInt32((int)queue_is_running); // true=1, false=0      
      (*&sendOSCMessage)(&local_mes);
       
   }

  }

  int getLength(){
    return queue_newest_index-queue_current_index;
  }

  int getCurrentIndex(){ // gibt den aktuellen index zurück
    return queue_current_index;
  }

  float getCurrentEntry(int data){    //Abrufen von Werten aus der Queue
    return queue[queue_current_index%queue_length][data];
  }

  void addToQueue(float new_mode, float new_value){ //Hinzufügen zu Queue
    queue[queue_newest_index%queue_length][0]=new_mode;
    queue[queue_newest_index%queue_length][1]=new_value;
    queue_newest_index++;
  }

  void removeFromQueue(){
    queue_current_index++;
  }

  void getQueueStart(OSCMessage *_mes) {
    setup();
    remote_length=_mes->getArgInt32(0); // länge des neuen blocks
  }
  
  void getQueueReset() {
    setup();
    remote_length=0;
  }
  
  void setQueuePause(boolean is_pause) {
    queue_is_pause=is_pause;
  }

  void sendQueueRequest() {
    OSCMessage local_mes;
    local_mes.beginMessage("/ml/q/req");
    local_mes.addArgInt32(remote_index);
    (*&sendOSCMessage)(&local_mes);
  }

  void getQueueResponse(OSCMessage *_mes) {

    float index=_mes->getArgInt32(0);
    float entry_mode=_mes->getArgFloat(1);
    float entry_value=_mes->getArgFloat(2);

    if (index==remote_index) { // alles richtig
      remote_index++;
      addToQueue(entry_mode,entry_value); 
   //   Serial.println("Index "+String(index)+" mode "+entry_mode+" Value "+entry_value);
    } 
    else {
   //   Serial.println("transfer error");
    }

  }
  
  void setQueueIsRunningFalse() {
    queue_is_running=false;
  }

};

